/* crlf -- convert line endings
   Copyright (C) 1999 Darren Salt.
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:
   1. Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.
   2. Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
   3. All advertising materials mentioning features or use of this software
      must display the following acknowledgement:
        This product includes software developed by Darren Salt.
   4. The name of the author may not be used to endorse or promote products
      derived from this software without specific prior written permission.

   This software is provided by the author "as is" and any express or implied
   warranties including, but not limited to, the implied warranties of
   merchantability and fitness for a particular purpose are disclaimed. in no
   event shall the author be liable for any direct, indirect, incidental,
   special, exemplary, or consequential damages (including, but not limited
   to, procurement of substitute goods or services; loss of use, data, or
   profits; or business interruption) however caused and on any theory of
   liability, whether in contract, strict liability, or tort (including
   negligence or otherwise) arising in any way out of the use of this
   software, even if advised of the possibility of such damage.
 */

/* Written by Darren Salt <ds@youmustbejoking.demon.co.uk>.  */

#include <system.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include <getopt.h>
#include <error.h>
#include <xalloc.h>

#ifdef __riscos
# include "ro_wild.h"
#endif

#include "ds_utils.h"

enum {
  type_lf, type_crlf, type_lfcr, type_cr
} endtype;


const char *program_name;

static const struct option opts[] =
{
  { "lf", no_argument, (int*)&endtype, type_lf },
  { "unix", no_argument, (int*)&endtype, type_lf },
  { "acorn", no_argument, (int*)&endtype, type_lf },

  { "crlf", no_argument, (int*)&endtype, type_crlf },
  { "dos", no_argument, (int*)&endtype, type_crlf },

  { "lfcr", no_argument, (int*)&endtype, type_lfcr },
  { "spool", no_argument, (int*)&endtype, type_lfcr },

  { "cr", no_argument, (int*)&endtype, type_cr },
  { "mac", no_argument, (int*)&endtype, type_lfcr },

  { "help", no_argument, 0, 'h' },
  { "version", no_argument, 0, 'v' }
};


static void
help (void)
{
  printf ("\
Usage: %s [OPTION] [FILE]...\n\
Concatenate FILE(s) or standard input to standard output, converting line\n\
endings.  LF, CRLF, LFCR and CR are recognised.\n\
The default conversion is to LF.\n\
\n\
  -c --crlf --dos           convert to CRLF\n\
  -C --cr   --mac           convert to CR\n\
  -l --lf   --unix --acorn  convert to LF\n\
  -L --lfcr --spool         convert to LFCR\n\
     --help                 display this help and exit\n\
     --version              output version information and exit\n\
", program_name);
  exit (0);
}


static void
version (void)
{
  printf ("crlf %s\n\
\n\
Copyright (C) 1999 Darren Salt.\n\
This is free software; see the source for copying conditions. There is NO\n\
warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n\
", Version);
  exit (0);
}


static void
writelf (void)
{
  switch (endtype)
    {
    case type_lf:
      putchar ('\n');
      break;
    case type_cr:
      putchar ('\r');
      break;
    case type_crlf:
      printf ("\r\n");
      break;
    case type_lfcr:
      printf ("\n\r");
    }
  if (ferror (stdout))
    error (2, errno, "write error");
}


#define BUFSIZE 8192
#define GETCHAR() \
  (ptr < lim ? buffer[ptr++] \
   : ((lim = fread (buffer, (ptr = 1), BUFSIZE, fd)) == 0 \
      ? (finished = EOF) : buffer[0]))


static int
crlf_file (const char *name)
{
  static char *buffer;
  int ptr = 0, lim = 0, finished = 0;
  int c, c2 = EOF;
  FILE *fd;

  if (!buffer)
    buffer = xmalloc (BUFSIZE);

  fd = (!strcmp (name, "-")) ? stdin : fopen (name, "r");
  if (!fd)
    {
      error (0, 0, "%s: couldn't open", name);
      return 1;
    }
  else if (fd == stdin)
    name = "(standard input)";
  while (!finished)
    {
      c = (c2 == EOF) ? GETCHAR () : c2;
      c2 = (c == EOF) ? EOF : GETCHAR ();
      if (ferror (fd))
	{
	  fclose (fd);
	  goto read_error;
	}
      switch (c)
	{
	case EOF:
	  return 0;
	case '\n':
	  writelf ();
	  if (c2 == '\r')
	    c2 = EOF;
	  break;
	case '\r':
	  writelf ();
	  if (c2 == '\n')
	    c2 = EOF;
	  break;
	default:
	  putchar (c);
	  if (!ferror (stdout))
	    break;
	  error (0, errno, "write error");
	  fclose (fd);
	  return 1;
	}
    }
  if (fd == stdin || !fclose (fd))
    return 0;
read_error:
  error (0, errno, "%s: read error", name);
  return 1;
}


int
main (int argc, char *const *argv)
{
  int status = 0, c;

  program_name = argv[0];

  while ((c = getopt_long (argc, argv, "cClL", opts, 0)) > 0)
    switch (c)
      {
      case 'c': endtype = type_cr; break;
      case 'C': endtype = type_crlf; break;
      case 'l': endtype = type_lf; break;
      case 'L': endtype = type_lfcr; break;
      case 'h': help (); break;
      case 'v': version ();
      }

  if (optind >= argc)
    return crlf_file ("-");
  else
    while (optind < argc)
      {
#ifdef __riscos
	if (ro_wild_test (argv[optind]))
	  {
	    char *name;
	    ro_wild *wild = ro_wild_init (argv[optind]);
	    while ((name = ro_wild_next (wild)) != 0)
	      status |= crlf_file (name);
	    if (!ro_wild_finish (wild))
	      {
		error (0, ENOENT, "%s", argv[optind]);
		status = 1;
	      }
	  }
	else
#endif
	  status |= crlf_file (argv[optind]);
	optind++;
      }
  return status;
}
